﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace HeuristicLab.Algorithms.DataAnalysis.Experimental {
  public static class CubicSplineGCV {
    // CUBGCV (X,F,DF,N,Y,C,IC,VAR,JOB,SE,WK,IER)
    //   ARGUMENTS    X      - VECTOR OF LENGTH N CONTAINING THE
    //                           ABSCISSAE OF THE N DATA POINTS
    //                           (X(I),F(I)) I=1..N. (INPUT) X
    //                           MUST BE ORDERED SO THAT
    //                           X(I) .LT. X(I+1).
    //                F      - VECTOR OF LENGTH N CONTAINING THE
    //                           ORDINATES (OR FUNCTION VALUES)
    //                           OF THE N DATA POINTS (INPUT).
    //                DF     - VECTOR OF LENGTH N. (INPUT/OUTPUT)
    //                           DF(I) IS THE RELATIVE STANDARD DEVIATION
    //                           OF THE ERROR ASSOCIATED WITH DATA POINT I.
    //                           EACH DF(I) MUST BE POSITIVE.  THE VALUES IN
    //                           DF ARE SCALED BY THE SUBROUTINE SO THAT
    //                           THEIR MEAN SQUARE VALUE IS 1, AND UNSCALED
    //                           AGAIN ON NORMAL EXIT.
    //                           THE MEAN SQUARE VALUE OF THE DF(I) IS RETURNED
    //                           IN WK(7) ON NORMAL EXIT.
    //                           IF THE ABSOLUTE STANDARD DEVIATIONS ARE KNOWN,
    //                           THESE SHOULD BE PROVIDED IN DF AND THE ERROR
    //                           VARIANCE PARAMETER VAR (SEE BELOW) SHOULD THEN
    //                           BE SET TO 1.
    //                           IF THE RELATIVE STANDARD DEVIATIONS ARE UNKNOWN,
    //                           SET EACH DF(I)=1.
    //                N      - NUMBER OF DATA POINTS (INPUT).
    //                           N MUST BE .GE. 3.
    //                Y,C    - SPLINE COEFFICIENTS. (OUTPUT) Y
    //                           IS A VECTOR OF LENGTH N. C IS
    //                           AN N-1 BY 3 MATRIX. THE VALUE
    //                           OF THE SPLINE APPROXIMATION AT T IS
    //                           S(T)=((C(I,3)*D+C(I,2))*D+C(I,1))*D+Y(I)
    //                           WHERE X(I).LE.T.LT.X(I+1) AND
    //                           D = T-X(I).
    //                IC     - ROW DIMENSION OF MATRIX C EXACTLY
    //                           AS SPECIFIED IN THE DIMENSION
    //                           STATEMENT IN THE CALLING PROGRAM. (INPUT)
    //                VAR    - ERROR VARIANCE. (INPUT/OUTPUT)
    //                           IF VAR IS NEGATIVE (I.E. UNKNOWN) THEN
    //                           THE SMOOTHING PARAMETER IS DETERMINED
    //                           BY MINIMIZING THE GENERALIZED CROSS VALIDATION
    //                           AND AN ESTIMATE OF THE ERROR VARIANCE IS
    //                           RETURNED IN VAR.
    //                           IF VAR IS NON-NEGATIVE (I.E. KNOWN) THEN THE
    //                           SMOOTHING PARAMETER IS DETERMINED TO MINIMIZE
    //                           AN ESTIMATE, WHICH DEPENDS ON VAR, OF THE TRUE
    //                           MEAN SQUARE ERROR, AND VAR IS UNCHANGED.
    //                           IN PARTICULAR, IF VAR IS ZERO, THEN AN
    //                           INTERPOLATING NATURAL CUBIC SPLINE IS CALCULATED.
    //                           VAR SHOULD BE SET TO 1 IF ABSOLUTE STANDARD
    //                           DEVIATIONS HAVE BEEN PROVIDED IN DF (SEE ABOVE).
    //                JOB    - JOB SELECTION PARAMETER. (INPUT)
    //                         JOB = 0 SHOULD BE SELECTED IF POINT STANDARD ERROR
    //                           ESTIMATES ARE NOT REQUIRED IN SE.
    //                         JOB = 1 SHOULD BE SELECTED IF POINT STANDARD ERROR
    //                           ESTIMATES ARE REQUIRED IN SE.
    //                SE     - VECTOR OF LENGTH N CONTAINING BAYESIAN STANDARD
    //                           ERROR ESTIMATES OF THE FITTED SPLINE VALUES IN Y.
    //                           SE IS NOT REFERENCED IF JOB=0. (OUTPUT)
    //                WK     - WORK VECTOR OF LENGTH 7*(N + 2). ON NORMAL EXIT THE
    //                           FIRST 7 VALUES OF WK ARE ASSIGNED AS FOLLOWS:-
    //
    //                           WK(1) = SMOOTHING PARAMETER (= RHO/(RHO + 1))
    //                           WK(2) = ESTIMATE OF THE NUMBER OF DEGREES OF
    //                                   FREEDOM OF THE RESIDUAL SUM OF SQUARES
    //                           WK(3) = GENERALIZED CROSS VALIDATION
    //                           WK(4) = MEAN SQUARE RESIDUAL
    //                           WK(5) = ESTIMATE OF THE TRUE MEAN SQUARE ERROR
    //                                   AT THE DATA POINTS
    //                           WK(6) = ESTIMATE OF THE ERROR VARIANCE
    //                           WK(7) = MEAN SQUARE VALUE OF THE DF(I)
    //
    //                           IF WK(1)=0 (RHO=0) AN INTERPOLATING NATURAL CUBIC
    //                           SPLINE HAS BEEN CALCULATED.
    //                           IF WK(1)=1 (RHO=INFINITE) A LEAST SQUARES
    //                           REGRESSION LINE HAS BEEN CALCULATED.
    //                           WK(2) IS AN ESTIMATE OF THE NUMBER OF DEGREES OF
    //                           FREEDOM OF THE RESIDUAL WHICH REDUCES TO THE
    //                           USUAL VALUE OF N-2 WHEN A LEAST SQUARES REGRESSION
    //                           LINE IS CALCULATED.
    //                           WK(3),WK(4),WK(5) ARE CALCULATED WITH THE DF(I)
    //                           SCALED TO HAVE MEAN SQUARE VALUE 1.  THE
    //                           UNSCALED VALUES OF WK(3),WK(4),WK(5) MAY BE
    //                           CALCULATED BY DIVIDING BY WK(7).
    //                           WK(6) COINCIDES WITH THE OUTPUT VALUE OF VAR IF
    //                           VAR IS NEGATIVE ON INPUT.  IT IS CALCULATED WITH
    //                           THE UNSCALED VALUES OF THE DF(I) TO FACILITATE
    //                           COMPARISONS WITH A PRIORI VARIANCE ESTIMATES.
    //
    //                IER    - ERROR PARAMETER. (OUTPUT)
    //                         TERMINAL ERROR
    //                           IER = 129, IC IS LESS THAN N-1.
    //                           IER = 130, N IS LESS THAN 3.
    //                           IER = 131, INPUT ABSCISSAE ARE NOT
    //                             ORDERED SO THAT X(I).LT.X(I+1).
    //                           IER = 132, DF(I) IS NOT POSITIVE FOR SOME I.
    //                           IER = 133, JOB IS NOT 0 OR 1.

    // TODO: Build x64 version of the library using mingw
    // TODO: statically link all required libraries (libgcc_s_dw2-1.dll is still required)
    // To build the fortran library use:
    // > gfortran -static -Lc:/mingw/bin/libgcc_s_dw2-1.dll -m32 642.f -shared -o 642_x86.dll
    // this means that libgcc_s_dw2-1 is statically linked, i386 code is produced and a shared library is produced
    [DllImport("642_x64.dll", CallingConvention = CallingConvention.Cdecl, EntryPoint = "cubgcv")]
    public static extern void cubgcv_x64(
      double[] x, double[] f, double[] df, ref int n, double[] y, double[,] c, ref int ic, ref double var, ref int job, double[] se, double[,] wk);

    // CUBGCV(X, F, DF, N, Y, C, IC, VAR, JOB, SE, WK, IER)
    [DllImport("642_x86.dll", CallingConvention = CallingConvention.Cdecl, EntryPoint = "cubgcv")]
    public static extern void cubgcv_x86(
      double[] x, 
      double[] f, 
      double[] df, 
      ref int n, 
      double[] y, 
      double[,] c, 
      ref int ic, 
      ref double var, 
      ref int job, 
      double[] se, 
      double[,] wk,
      ref int ier);
  }

}
